//! Rocks DB implementation of the storage layer
use crate::{event_filter::EventFilter, RocksDb};
use futures::{Future, FutureExt, Stream};
use nostr_rs_storage_base::{Error, Storage};
use nostr_rs_types::types::Event;
use rocksdb::{BoundColumnFamily, DBIteratorWithThreadMode, DB};
use std::{
    collections::VecDeque,
    pin::Pin,
    sync::Arc,
    task::{Context, Poll},
};

type CurrentEventByPrefixFuture<'a> = Pin<
    Box<
        dyn Future<
                Output = Result<Option<nostr_rs_types::types::Event>, nostr_rs_storage_base::Error>,
            > + Send
            + 'a,
    >,
>;

pub struct WrapperIterator<'a> {
    /// Reference to the rocks db database. This is useful to load the event
    /// data, because in the secondary indexes we only store the event ID. It
    /// could be possible to avoid a reference and speed up things even more to
    /// duplicate the event, but that would mean it would use way more disk
    /// space.
    pub db: &'a RocksDb,
    /// List of filters to apply to the events, before returning. If no filter
    /// is given each events from the secondary index will be returned,
    /// otherwise the events will be filtered by the given filter, and only
    /// those events that comply will be returned
    pub filter: Option<EventFilter>,
    /// Reference to the namespace to use to query the secondary index. If none
    /// is given the secondary_index_iterator must be constructed outside this
    /// wrapper.
    pub namespace: Option<Arc<BoundColumnFamily<'a>>>,
    /// The current secondary index iterator. If none is given the iterator will
    /// try to create one using the namespace property and the first prefix from
    /// prefixes (it will also be copied to current_prefix)
    pub secondary_index_iterator: Option<DBIteratorWithThreadMode<'a, DB>>,
    pub current_prefix: Vec<u8>,
    pub prefixes: VecDeque<Vec<u8>>,
    pub limit: Option<usize>,
    pub returned: usize,

    pub current_event_by_prefix: Option<CurrentEventByPrefixFuture<'a>>,
}

impl<'a> WrapperIterator<'a> {
    /// Selects the next prefix available and starts an iterator using the
    /// secondary index. If no prefix is available from prefixes the functions
    /// return None, signalling upstream the are no more results
    fn select_next_prefix_using_secondary_index(&mut self) -> Option<()> {
        let prefix = self.prefixes.pop_front()?;
        self.secondary_index_iterator = Some(
            self.db
                .db
                .prefix_iterator_cf(self.namespace.as_ref()?, prefix.clone()),
        );
        self.current_prefix = prefix;
        Some(())
    }
}

impl<'a> Stream for WrapperIterator<'a> {
    type Item = Result<Event, Error>;

    fn size_hint(&self) -> (usize, Option<usize>) {
        (0, None)
    }

    fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        if Some(self.returned) == self.limit {
            return Poll::Ready(None);
        }

        let this = Pin::into_inner(self);
        let db = this.db;

        if let Some(mut current_event_filter) = this.current_event_by_prefix.take() {
            match current_event_filter.poll_unpin(cx) {
                Poll::Ready(Ok(Some(event))) => {
                    // event is ready, apply the neccesary filters
                    if let Some(filter) = &this.filter {
                        if filter.check_event(&event) {
                            this.returned += 1;
                            return Poll::Ready(Some(Ok(event)));
                        }
                    } else {
                        this.returned += 1;
                        return Poll::Ready(Some(Ok(event)));
                    }
                }
                Poll::Ready(Err(x)) => return Poll::Ready(Some(Err(x))),
                Poll::Pending => {
                    // add it back
                    this.current_event_by_prefix = Some(current_event_filter);
                    return Poll::Pending;
                }
                _ => {}
            }
        }
        let secondary_index = if let Some(iterator) = this.secondary_index_iterator.as_mut() {
            iterator
        } else {
            return Poll::Ready(None);
        };

        match secondary_index.next() {
            Some(Ok((key, value))) => {
                if !key.starts_with(&this.current_prefix) {
                    if this.select_next_prefix_using_secondary_index().is_none() {
                        return Poll::Ready(None);
                    }
                } else {
                    // query the database to get the record
                    this.current_event_by_prefix = Some(db.get_event(value));
                }

                Poll::Pending
            }
            Some(Err(err)) => Poll::Ready(Some(Err(Error::Internal(err.to_string())))),
            None => {
                if this.namespace.is_some() {
                    if this.select_next_prefix_using_secondary_index().is_none() {
                        return Poll::Ready(None);
                    }
                } else {
                    // No secondary index is used to query, this means the query is
                    // using the ID filter, so it is more efficient to use the
                    // primary index to prefetch events that may satisfy the query
                    let current_event_by_prefix =
                        this.prefixes.pop_front().map(|prefix| db.get_event(prefix));
                    this.current_event_by_prefix = current_event_by_prefix;
                }
                Poll::Pending
            }
        }
    }
}
